# -*- coding: utf-8 -*-
"""
Created on Sat Feb 16 07:18:29 2019

@author: limingfan
"""

import tensorflow as tf

from Zeras.nn import get_position_emb_mat, get_emb_positioned
from Zeras.nn import get_tensor_expanded
from Zeras.nn import get_mask_mat_subsequent
from Zeras.nn import get_label_smoothened

from model_encoder_decoder import ModelEncoderDecoder, Generator
from model_encoder_decoder import Encoder, Decoder


class ModelGraph():
    
    @staticmethod
    def build_placeholder(settings):
        
        src_seq = tf.placeholder(tf.int32, [None, None], name='src_seq')  # id in vocab
        src_seq_mask = tf.placeholder(tf.int32, [None, None], name='src_seq_mask')
        
        dcd_seq = tf.placeholder(tf.int32, [None, None], name='dcd_seq')  # id in vocab
        dcd_seq_mask = tf.placeholder(tf.int32, [None, None], name='dcd_seq_mask')

        labels_seq = tf.placeholder(tf.int32, [None, None], name='labels_seq')  # id in vocab
        labels_mask = tf.placeholder(tf.int32, [None, None], name='labels_mask')
        
        # input sequence: could not prefix and suffix, when preparing examples
        # label sequence: suffix with a [end] token, then do [pad].
        #
        # decoder input seq: prefix with [start], suffix with [end], then do [pad].
        #
        print(src_seq)
        #
        input_tensors = (src_seq, src_seq_mask, dcd_seq, dcd_seq_mask)
        label_tensors = (labels_seq, labels_mask)
        #
        return input_tensors, label_tensors
    
    @staticmethod
    def build_inference(settings, input_tensors):
        
        src_seq, src_seq_mask, dcd_seq, dcd_seq_mask = input_tensors
        
        #
        dim_all = settings.num_heads * settings.num_units
        #
        keep_prob = tf.get_variable("keep_prob", shape=[], dtype=tf.float32, trainable=False)
        #
        with tf.device('/cpu:0'):
            emb_mat = tf.get_variable('token_embeddings',
                                      [settings.vocab.size(), settings.vocab.emb_dim],
                                      initializer=tf.constant_initializer(settings.vocab.embeddings),
                                      trainable = settings.emb_tune)
        #
        pe_mat = get_position_emb_mat(settings.max_seq_len, settings.posi_emb_dim,
                                      settings.dim_model, "posi_embeddings")
        #
        with tf.variable_scope("encoder_decoder"):

            emb_trans = lambda x: get_emb_positioned(x, emb_mat, pe_mat)
            
            encoder = Encoder(emb_trans, keep_prob, settings)
            decoder = Decoder(emb_trans, keep_prob, settings)
            generator = Generator(dim_all, settings.vocab.size(), emb_mat=emb_mat)
            
            model = ModelEncoderDecoder(encoder, decoder, generator)
            #
            # model vars are all defined by now
            # graph yet
            #
            
        #
        src_mask = get_tensor_expanded(src_seq_mask, 1, dtype=tf.float32)
        crs_mask = get_tensor_expanded(src_seq_mask, 1, dtype=tf.float32)
        dcd_mask = get_tensor_expanded(dcd_seq_mask, 1, dtype=tf.float32)
        sub_mask = get_mask_mat_subsequent(settings.max_len_decoding)
        dcd_mask = dcd_mask * sub_mask
        #
        if settings.is_train:
            out = model.forward(src_seq, src_mask, dcd_seq, dcd_mask, crs_mask)
            logits = model.generator.forward(out)
            logits_normed = tf.nn.softmax(logits, -1, name = 'logits')
            preds = tf.argmax(logits, -1, name="preds")
        else:
            if settings.beam_width == 1:
                logits, preds_d = model.do_greedy_decoding(src_seq, src_mask,
                                                           settings.max_len_decoding,
                                                           sub_mask, crs_mask,
                                                           settings.start_symbol_id)
                logits_normed = tf.identity(logits, name = 'logits')
                preds = tf.identity(preds_d, name="preds")
            else:
                logits, preds_d = model.do_beam_search_decoding(src_seq, src_mask,
                                                                settings.max_len_decoding,
                                                                sub_mask, crs_mask,
                                                                settings.start_symbol_id,
                                                                settings.beam_width)
                logits_normed = tf.identity(logits, name = 'logits')
                preds = tf.identity(preds_d, name="preds")
            #
        
        #
        print(logits_normed)
        print(preds)
        #
        output_tensors = logits_normed, logits, preds
        #   
        return output_tensors
    
    @staticmethod
    def build_loss_and_metric(settings, output_tensors, label_tensors):
        
        normed_logits, logits, preds = output_tensors
        labels_seq, labels_mask = label_tensors
        
        labels_len = tf.cast(tf.reduce_sum(labels_mask, -1), dtype=tf.float32)
        
        with tf.variable_scope('loss'):
            #
            onehot_labels = tf.one_hot(labels_seq, settings.vocab.size())
            labels_smooth = get_label_smoothened(onehot_labels, settings.vocab.size(),
                                                 settings.label_smoothing)
            #
            cross_ent = tf.nn.softmax_cross_entropy_with_logits(labels = labels_smooth,
                                                                logits = logits)
            loss_batch = tf.multiply(cross_ent, tf.cast(labels_mask, dtype=tf.float32))
            loss_batch = tf.reduce_sum(loss_batch, -1)
            loss_batch = tf.divide(loss_batch, labels_len, name='loss_batch')
            loss = tf.reduce_mean(loss_batch, axis = 0, name = 'loss')
        
        with tf.variable_scope('metric'):
            
            # correct_pred = tf.cast(tf.equal(labels_seq, preds), tf.int32)
            # acc = tf.reduce_mean(tf.cast(correct_pred, tf.float32), name = 'metric')
            
            metric = tf.constant(0.1, name = 'metric')
            
        #
        print(loss)
        print(metric)
        #
        return loss, metric
        #

        


# -*- coding: utf-8 -*-
"""
Created on Sat Sep  1 17:14:19 2018

@author: limingfan
"""

import numpy as np
import tensorflow as tf

#
def get_tensor_expanded(x, dim, dtype=None):
    """
    """
    x = tf.expand_dims(x, dim)
    if dtype is not None:
        x = tf.cast(x, dtype=dtype)
    #
    return x

#
def get_mask_mat_subsequent(size, name="mask_subsequent"):
    """ subsequent mask
    """
    mask_tensor = tf.constant(1.0, shape = (1, size, size), dtype=tf.float32)
    mask_tensor = tf.linalg.band_part(mask_tensor,
                                      num_lower = -1,
                                      num_upper = 0,
                                      name = name)
    return mask_tensor
    
#
def get_position_emb_mat(max_seq_len, posi_emb_dim, posi_emb_model,
                         name="position_embeddings"):
    """
    """
    d_model_recip_2 = 2.0 / posi_emb_model
    
    arg_mat = np.zeros((max_seq_len, posi_emb_dim), dtype=np.float32)
    for idx in range(max_seq_len):
        for idm in range(posi_emb_dim):
            arg_mat[idx, idm] = idx * 1e-4**(d_model_recip_2 * idm)
    #
    pe_sin = np.sin(arg_mat)
    pe_cos = np.cos(arg_mat)
    #    
    pe_sin = np.expand_dims(pe_sin, -1)  
    pe_cos = np.expand_dims(pe_cos, -1)
    pe_all = np.concatenate([pe_sin, pe_cos], -1)  # (T, D, 2)
    #
    pe_all = np.reshape(pe_all, [max_seq_len, -1])
    pe_all = pe_all[:, 0:posi_emb_dim]
    
    #
    # tf.Tensor
    pe_mat = tf.get_variable(name, shape = (max_seq_len, posi_emb_dim),
                             initializer = tf.constant_initializer(pe_all),
                             trainable = False)
        
    return pe_mat

def get_emb_positioned(x, token_emb, position_emb):
    """ x: [None, None]
    """
    posi = tf.range(tf.shape(x)[-1])
    
    seq_emb_t = tf.nn.embedding_lookup(token_emb, x)
    seq_emb_p = tf.nn.embedding_lookup(position_emb, posi)
    
    return seq_emb_t + seq_emb_p

#
def gelu(x):
    cdf = 0.5 * (1.0 + tf.tanh((0.79788456 * (x + 0.044715 * tf.pow(x, 3)) )))
    return x * cdf

#
def get_label_smoothened(onehot_label, num_classes, delta):
    """
    """
    new_label = (1.0 - delta) * onehot_label + delta / num_classes
    return new_label
    
#



# -*- coding: utf-8 -*-
"""
Created on Sat Feb 16 07:18:29 2019

@author: limingfan
"""

# import tensorflow as tf


class ModelGraphTemplate():
    """ Template for model graph.
        Three static methods
    """    
    @staticmethod
    def build_placeholder(settings):
        """
        """        
        # src_seq = tf.placeholder(tf.int32, [None, None], name='src_seq')  # id in vocab
        # src_seq_mask = tf.placeholder(tf.int32, [None, None], name='src_seq_mask')
        
        # dcd_seq = tf.placeholder(tf.int32, [None, None], name='dcd_seq')  # id in vocab
        # dcd_seq_mask = tf.placeholder(tf.int32, [None, None], name='dcd_seq_mask')

        # labels_seq = tf.placeholder(tf.int32, [None, None], name='labels_seq')  # id in vocab
        # labels_mask = tf.placeholder(tf.int32, [None, None], name='labels_mask')
        
        #
        # input sequence: could not prefix and suffix, when preparing examples
        # label sequence: suffix with a [end] token, then do [pad].
        #
        # decoder input seq: prefix with [start], suffix with [end], then do [pad].
        #
        # print(src_seq)
        #
        # input_tensors = (src_seq, src_seq_mask, dcd_seq, dcd_seq_mask)
        # label_tensors = (labels_seq, labels_mask)
        input_tensors = None
        label_tensors = None
        #
        return input_tensors, label_tensors
    
    #
    @staticmethod
    def build_inference(settings, input_tensors):
        """
        """
        #
        # print(logits_normed)
        # print(preds)
        #
        # output_tensors = logits_normed, logits, preds
        output_tensors = None
        #   
        return output_tensors
    
    #
    @staticmethod
    def build_loss_and_metric(settings, output_tensors, label_tensors):
        """
        """            
        #
        # print(loss)
        # print(metric)
        #
        loss = None
        metric = None
        #
        return loss, metric
        #



# -*- coding: utf-8 -*-
"""
Created on Sun Apr  7 21:01:52 2019

@author: limingfan
"""

import tensorflow as tf

#
def do_greedy_decoding(model, src, src_mask, max_len,
                       sub_mask, crs_mask, start_symbol_id):
    """
    """
    memory = model.encode(src, src_mask)
    
    batch_start = tf.cast(tf.ones_like(src[:, 0:1]), dtype = tf.int64)
    batch_start = tf.multiply(batch_start, start_symbol_id)    
    pad_tokens = tf.cast(tf.zeros_like(src[:, 0:1]), dtype = tf.int64)
    pad_tokens = tf.tile(pad_tokens, [1, max_len-1])
    
    logits_list = []
    preds_list = []
    
    dcd_feed = tf.concat([batch_start, pad_tokens], 1)
    for step in range(max_len):        
        out = model.decode(dcd_feed, sub_mask, memory, crs_mask)
        out_last = out[:, step, :]
        logits_curr = model.generator.forward(out_last)
        logits_curr = tf.nn.softmax(logits_curr, -1)
        preds_curr = tf.argmax(logits_curr, -1)
        
        logits_list.append(logits_curr)
        preds_list.append(preds_curr)
        #
        preds = tf.transpose(tf.stack(preds_list, 0), [1, 0])
        dcd_feed = tf.concat([batch_start, preds, pad_tokens[:, step+1:]], 1)
        #
    #
    logits = tf.transpose(tf.stack(logits_list, 0), [1, 0, 2])
    return logits, preds


def do_beam_search_decoding(model, src, src_mask, max_len,
                            sub_mask, crs_mask, start_symbol_id, beam_width):
    """
    """
    
    pass




# -*- coding: utf-8 -*-
"""
Created on Tue Aug 28 21:12:23 2018

@author: limingfan
"""

from Zeras.model_settings_baseboard import ModelSettingsBaseboard


class ModelSettings(ModelSettingsBaseboard):
    """
    """
    def __init__(self):
        """
        """
        super(ModelSettings, self).__init__()
        
        # task
        self.task = "copy"
        self.tokens_file = "./vocab/vocab_tokens.txt"
        self.emb_file = None
        
        # model
        self.model_tag = None
        self.is_train = None

        # data macro
        self.min_seq_len = 1     #
        self.max_seq_len = 12    #
        
        # vocab
        self.emb_dim = 128
        self.emb_tune = 1  # 1 for tune, 0 for not
        self.posi_emb_dim = self.emb_dim

        # model macro
        self.num_layers = 6
        self.num_heads = 8
        self.num_units = int(self.emb_dim / self.num_heads)
        self.dim_all = self.num_heads * self.num_units
        self.dim_model = self.emb_dim
        self.dim_ffm = 256
        #
        
        # self.decoder_vocab_size = 17
        self.beam_width = 1
        self.max_len_decoding = self.max_seq_len
        self.start_symbol_id = 2  #
        
        #
        # train
        self.gpu_available = "0"  # specified in args
        self.gpu_batch_split = [16, 20]   # list; if None, batch split evenly
        #
        self.gpu_mem_growth = True
        self.log_device = False
        self.soft_placement = True
        
        self.with_bucket = False
        
        #
        self.num_epochs = 100
        self.batch_size = 36
        self.batch_size_eval = 6
        self.max_batches_eval = 20
        
        self.reg_lambda = 0.0  # 0.0, 0.01
        self.grad_clip = 0.0  # 0.0, 5.0, 8.0, 2.0
        self.keep_prob = 0.9  # 1.0, 0.7, 0.5
        self.label_smoothing = 0.01
        
        self.optimizer_type = 'adam'  # adam, momentum, sgd, customized
        self.momentum = 0.9
        self.learning_rate_base = 0.001   #
        self.learning_rate_minimum = 0.000001
        self.warmup_steps = 1000
        self.decay_steps = 5000
        self.decay_rate = 0.99
        self.staircase = True
        
        self.check_period_batch = 100
        self.valid_period_batch = 100
        #

        # inputs/outputs
        self.vs_str_multi_gpu = "vs_multi_gpu"
        #
        self.inputs_predict_name = ['src_seq:0', 'src_seq_mask:0']
        self.outputs_predict_name = ['vs_multi_gpu/logits:0']
        self.pb_outputs_name = ['vs_multi_gpu/logits']
                
        self.inputs_train_name = ['src_seq:0', 'src_seq_mask:0',
                                  'dcd_seq:0', 'dcd_seq_mask:0',
                                  'labels_seq:0', 'labels_mask:0']
        self.outputs_train_name = ['vs_multi_gpu/logits:0']
        self.use_metric = True
        
        self.debug_tensors_name = ['vs_multi_gpu/loss/loss:0',
                                   'vs_multi_gpu/logits:0',
                                   'vs_multi_gpu/preds:0'
                                   #'encoder/encoder_rnn_1/bw/bw/sequence_length:0',
                                   #'inputs_len:0'
                                   ]
        
        #
        self.base_dir = './task_copy_results'
        # self.model_dir = None   # if not set, default values will be used.
        # self.model_name = None
        # self.pb_file = None
        # self.log_dir = None
        # self.log_path = None
        #
   
#      
if __name__ == "__main__":
    
    sett = ModelSettings()
    #
    sett.model_tag = 'cnn'
    sett.is_train = False
    #
    sett.check_settings()
    #
    
    #
    info_dict = sett.trans_info_to_dict()
    print("original:")
    print(info_dict)
    print()
    #
    info_dict["model_tag"] = "transformer"
    sett.assign_info_from_dict(info_dict)
    #
    info_dict = sett.trans_info_to_dict()
    print("assigned:")
    print(info_dict)
    print()
    #
    
    #
    file_path = "./temp_settings.json"
    sett.save_to_json_file(file_path)    
    sett.load_from_json_file(file_path)
    #
    info_dict = sett.trans_info_to_dict()
    print("saved then loaded:")
    print(info_dict)
    print()
    #

    #    
    sett.close_logger()
    #


# -*- coding:utf8 -*-

import numpy as np
import struct
import io


class Vocab(object):
    
    pad_token = '[pad]'
    unk_token = '[unk]'
    start_token = '[start]'
    end_token = '[end]'
    mask_token = '[mask]'
    replace_token = '[replace]'
    unchange_token = '[unchge]'
    vocab_start_idx = 7
    
    delimiter_str = '[strsep]'
    
    def __init__(self, initial_tokens = [], lower = False):
        """
        """
        self.dict_id2token = {}
        self.dict_token2id = {}
        self.dict_token_cnt = {}
        self.lower = lower

        self.emb_dim = 64
        self.embeddings = None

        self._add_predefined_tokens()
        
        self.initial_tokens = initial_tokens
        for token in self.initial_tokens:
            self.add(token, cnt = 10000)
            
    def _add_predefined_tokens(self):
        
        self.add(self.pad_token, 10000)  # make pad_token id: 0
        self.add(self.unk_token, 10000)
        self.add(self.start_token, 10000)
        self.add(self.end_token, 10000)
        self.add(self.mask_token, 10000)
        self.add(self.replace_token, 10000)   
        self.add(self.unchange_token, 10000)
        #

    def size(self):
        return len(self.dict_id2token)

    def get_id(self, token):
        token = token.lower() if self.lower else token
        try:
            return self.dict_token2id[token]
        except KeyError:
            return self.dict_token2id[self.unk_token]

    def get_token(self, idx):
        try:
            return self.dict_id2token[idx]
        except KeyError:
            return self.unk_token

    def add(self, token, cnt=1):
        #
        token = token.lower() if self.lower else token
        if token in self.dict_token2id:
            idx = self.dict_token2id[token]
            # print('same token: %s' % token)
        else:
            idx = len(self.dict_id2token)
            self.dict_id2token[idx] = token
            self.dict_token2id[token] = idx
        if cnt > 0:
            if token in self.dict_token_cnt:
                self.dict_token_cnt[token] += cnt
            else:
                self.dict_token_cnt[token] = cnt
        return idx
        
    def add_tokens_from_vocab(self, vocab):
        """ add tokens from another vocab instance
        """
        for token in vocab.dict_token_cnt:
            self.add(token, vocab.dict_token_cnt[token])
        
    def add_tokens_from_corpus(self, corp):
        """ add tokens from corpus (list)
            with each item in the list as a list of tokens 
        """
        for item in corp:
            for word in item:
                self.add(word)
                
    def add_tokens_from_file(self, file_path):
        """ add tokens from file
            with one token in one line
        """
        fp = open(file_path, 'r', encoding='utf-8')
        lines = fp.readlines()
        fp.close()
        self.add_tokens_from_lines(lines)
        
    def add_tokens_from_lines(self, lines):
        """ add tokens from lines
            with one token in one line
        """
        token = ' '      
        for idx, line in enumerate(lines):
            if line.startswith(' '):
                self.add(' ', 10000)
                print('WARNING: blank token')
                continue
            #            
            str_arr = line.strip().split()  #
            # print(str_arr)
            if len(str_arr) == 0:
                print('WARNING: len(str_arr) == 0 with prev token: %s' % token)
                # continue
            #
            token = str_arr[0]      # .strip()
            # token = line.rstrip('\n')
            # print(token)
            if len(str_arr) > 1:
                cnt = int(str_arr[1].strip())
                self.add(token, cnt)
            else:
                self.add(token)
        #
        print('num lines: %d' % (idx + 1) )
        print('num tokens after loading: %d' % len(self.dict_id2token))
        #
        
    def save_tokens_to_file(self, file_path):
        """ save tokens to file
            with one token in one line
        """
        with open(file_path, 'w', encoding='utf-8') as fp:
            for tid in range(self.size()):
                token = self.dict_id2token[tid]
                fp.write(token + ' ' + str(self.dict_token_cnt[token]) + '\n')
    
    #
    def filter_tokens_by_cnt(self, min_cnt):
        #
        #filtered_tokens = [token for token in self.token2id if self.token_cnt[token] >= min_cnt]
        filtered_tokens = [self.dict_id2token[idd] for idd in range(self.size())
                                   if self.dict_token_cnt[self.dict_id2token[idd]] >= min_cnt]
        # rebuild the token ~ id map
        self.dict_token2id = {}
        self.dict_id2token = {}
        self._add_predefined_tokens()
        for token in self.initial_tokens:
            self.add(token, cnt=0)
        for token in filtered_tokens:
            self.add(token, cnt=0)
        
    #
    def randomly_init_embeddings(self, emb_dim):
        #
        self.emb_dim = emb_dim
        self.embeddings = np.random.rand(self.size(), emb_dim)
        #for token in [self.pad_token, self.unk_token]:
        #for token in [self.pad_token]: self.embeddings[self.get_id(token)] = np.zeros([self.emb_dim])
        
    def load_pretrained_embeddings(self, emb_path, load_existing = True):
        #
        if emb_path is None:
            self.randomly_init_embeddings(self.emb_dim)
            return
        #
        if emb_path.endswith('bin'):
            self._load_pretrained_embeddings_bin(emb_path, load_existing)
        elif emb_path.endswith('txt'):
            self._load_pretrained_embeddings_txt(emb_path, load_existing)
        #
        
    def _load_pretrained_embeddings_bin(self, emb_path, load_existing = True):

        num_tokens = 0
        emb_dim = 0
        trained_embeddings = {}
        with open(emb_path, 'rb') as fp:
            str_t = ''
            b, = struct.unpack('b', fp.read(1))
            while b != 32:
                str_t += chr(b)
                b, = struct.unpack('b', fp.read(1))
            #            
            num_tokens = int(str_t)
            # print(str_t)
            #
            str_t = ''
            b, = struct.unpack('b', fp.read(1))
            while b != 10:
                str_t += chr(b)
                b, = struct.unpack('b', fp.read(1))
            #
            emb_dim = int(str_t)
            # print(str_t)
            #
            for idx in range(num_tokens):
                #
                count = 0
                b, = struct.unpack('b', fp.read(1))
                while b != 32:
                    count += 1
                    b, = struct.unpack('b', fp.read(1))
                #
                fp.seek(-count-1, io.SEEK_CUR)
                data = fp.read(count)
                token = data.decode('utf-8')
                # print(data)
                # print(token)
                fp.read(1)
                #
                emb_list = []
                for d in range(emb_dim):
                    data = fp.read(4)
                    elem, = struct.unpack("f", data)
                    emb_list.append(elem)
                fp.read(1)
                #
                trained_embeddings[token] = emb_list
                #
        #
        self.emb_dim = emb_dim
        #
        # rebuild the token x id map
        #
        # initiate embeddings
        self.embeddings = np.random.rand(self.size(), self.emb_dim)
        # for token in [self.pad_token]: self.embeddings[self.get_id(token)] = np.zeros([self.emb_dim])
        # self.embeddings = np.zeros([self.size(), self.emb_dim])
        #
        # load embeddings        
        for token in self.dict_token2id.keys():
            if token in trained_embeddings:
                self.embeddings[self.get_id(token)] = trained_embeddings[token]
    
    def _load_pretrained_embeddings_txt(self, emb_path, load_existing = True):
        #
        trained_embeddings = {}
        with open(emb_path, 'r', encoding='utf-8') as fin:
            for line in fin:
                contents = line.strip().split()
                #
                if len(contents) < 3: continue  #
                #
                if line.startswith(' '):
                    token = ' '
                    trained_embeddings[token] = list(map(float, contents[0:]))
                    continue
                #                
                token = contents[0]
                # print(token)
                #
                if load_existing and token not in self.dict_token2id:
                    continue  # only existing tokens
                #
                trained_embeddings[token] = list(map(float, contents[1:]))
                #
        #
        trained_tokens = list(trained_embeddings.keys() )
        #if self.emb_dim is None:
        self.emb_dim = len(trained_embeddings[trained_tokens[0]])
        #
        # rebuild the token x id map
        #
        # initiate embeddings
        self.embeddings = np.random.rand(self.size(), self.emb_dim)
        # for token in [self.pad_token]: self.embeddings[self.get_id(token)] = np.zeros([self.emb_dim])
        # self.embeddings = np.zeros([self.size(), self.emb_dim])
        #
        # load embeddings        
        for token in self.dict_token2id.keys():
            if token in trained_embeddings:
                self.embeddings[self.get_id(token)] = trained_embeddings[token]
                
    def save_embeddings_to_file(self, emb_path):
        #
        if emb_path.endswith('bin'):
            self._save_embeddings_to_file_bin(emb_path)
        elif emb_path.endswith('txt'):
            self._save_embeddings_to_file_txt(emb_path)
        #
        
    def _save_embeddings_to_file_bin(self, emb_path):
        
        num_words_str = str(self.size()) + ' '
        emb_dim_str = str(self.emb_dim) + '\n'
        
        with open(emb_path, 'wb') as fp:
            #
            fp.write(num_words_str.encode('utf-8'))  # str to bytes
            fp.write(emb_dim_str.encode('utf-8'))
            #
            for idx in range(self.size()):
                word_and_space = self.dict_id2token[idx] + ' '
                fp.write(bytes(word_and_space, encoding = "utf-8"))  # str to bytes
                
                emb_value = self.embeddings[idx]
                for item in emb_value:
                    write_buf = struct.pack('f', item)   # float to bytes
                    fp.write(write_buf)
                #
                a = struct.pack('B', ord('\n') )   # char to byte
                fp.write(a)
                #
                
    def _save_embeddings_to_file_txt(self, emb_path):
        #
        with open(emb_path, 'w', encoding = 'utf-8') as fp:
            for idd in range(self.size()):
                token = self.dict_id2token[idd]
                emb_str = map(str, self.embeddings[idd])
                line = token + ' ' + ' '.join(emb_str) + '\n'
                fp.write(line)

    def convert_tokens_to_ids(self, tokens):
        """ convert a list of tokens to a list of ids,
            tokens: a list of tokens
        """
        vec = [self.get_id(label) for label in tokens]
        return vec

    def convert_ids_to_tokens(self, ids, stop_id = None):
        """ convert a list of ids to a list of tokens,
            break when stop_id is turning up in ids,
            ids: a list of ids
        """
        tokens = []
        for i in ids:            
            if stop_id is not None and i == stop_id:
                break
            tokens += [self.get_token(i)]
        return tokens

#
if __name__ == '__main__':
    
    import time
    
    vocab = Vocab()
    
    vocab.add_tokens_from_file('./vocab/vocab_tokens.txt')
    print('tokens loaded from vocab_tokens.txt')
    print(vocab.size())
    
    s = time.time()
    vocab.load_pretrained_embeddings('./vocab/vocab_emb.bin')
    e = time.time()
    print('emb loaded from vocab_emb.bin')
    print('time cost: %g' % (e-s))
    
    vocab.save_embeddings_to_file('./vocab/vocab_emb.txt')
    print('emb saved to vocab_emb.txt')
    
    #
    vocab.add_tokens_from_file('./vocab/vocab_emb.txt')
    print('tokens loaded from vocab_emb.txt')
    print(vocab.size())
    print()
    
    vocab.save_tokens_to_file('./vocab/vocab_tokens.txt')
    print('tokens saved to vocab_tokens.txt')
    print()
    
    # print(vocab.embeddings[vocab.get_id(' ')])
    # print()
    
    #
    s = time.time()
    vocab.load_pretrained_embeddings('./vocab/vocab_emb.txt')
    e = time.time()    
    print('emb loaded from vocab_emb.txt')
    print('time cost: %g' % (e-s))
    print(vocab.size())
    print(vocab.emb_dim)
    
    # print(vocab.embeddings[vocab.get_id(' ')])
    # print()

    vocab.save_embeddings_to_file('./vocab/vocab_emb.bin')
    print('emb saved to vocab_emb.bin')
    
    
    #
    vocab.add_tokens_from_file('./vocab/vocab_tokens.txt')
    print('tokens loaded from vocab_tokens.txt')
    print(vocab.size())
    
    print(vocab.embeddings[vocab.get_id(' ')])
    print()
    
      


# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
# Modifications Copyright 2017 Abigail See
# Modifications Copyright 2018 Ming-Fan Li
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import time
import random

import queue as Queue

# from multiprocessing import Process        # True Process
# from multiprocessing.dummy import Process    # A wrapper of Thread

def get_generator_for_list(list_data):
    """
    """
    def gen(single_pass=True):
        while True:
            for item in list_data: yield item
            if single_pass: break
            #
    return gen

#
class DataBatcher(object):    
    """ This class is meant to be task-agnostic
    """    
    BATCH_QUEUE_MAX = 300       # max number of batches the batch_queue can hold
    BATCH_TIME_OUT = 5          # seconds
    EXAMPLE_TIME_OUT = 2
    
    def __init__(self, example_gen_or_list, batch_standardizer,
                 batch_size, single_pass, with_bucket=False, worker_type="thread"):
        """
        """
        if isinstance(example_gen_or_list, list):
            self.example_gen = get_generator_for_list(example_gen_or_list)
        else:
            self.example_gen = example_gen_or_list
        #
        self.batch_standardizer = batch_standardizer
        self.batch_size = batch_size
        self.with_bucket = with_bucket

        # queue
        self.single_pass = single_pass
        if single_pass:
            self.num_example_q_workers = 1 # one worker, so we read through the dataset just once
            self.num_batch_q_workers = 3   # num_workers to batch examples
            self.bucketing_cache_size = 1  # this essentially means no bucketing
        else:
            self.num_example_q_workers = 12 # num workers to fill example queue
            self.num_batch_q_workers = 12   # num workers to fill batch queue
            self.bucketing_cache_size = 200 # how many batches-worth of examples to load into cache before bucketing
        
        # worker_type
        self.worker_type = worker_type   # "process", "thread"
        #
        self.build_queue_and_workers(self.worker_type)
        #
        
    #
    def get_next_batch(self):
        """
        """        
        if self._batch_queue.qsize() == 0:
            print('batch_q_size: %i, example_q_size: %i' % (
                  self._batch_queue.qsize(), self._example_queue.qsize()) )
        #
        try:
            batch = self._batch_queue.get(timeout = self.BATCH_TIME_OUT) # get the next Batch
            return batch    
        except BaseException:
            return None
        
    #
    def build_queue_and_workers(self, worker_type="thread"):
        """
        """
        self.worker_type = worker_type
        #
        if worker_type == "process":
            from multiprocessing import Process        # True Process
        else:
            from multiprocessing.dummy import Process    # A wrapper of Thread
        #
        self.Process = Process
        #
        
        # queue
        self._batch_queue = Queue.Queue(self.BATCH_QUEUE_MAX)
        self._example_queue = Queue.Queue(self.BATCH_QUEUE_MAX * self.batch_size)
        
        self._finished_reading = False
        self.count_put_examples = 0
        self.count_get_examples = 0 
        #
        # workers
        self._example_q_workers = []        
        for _ in range(self.num_example_q_workers):
            self._example_q_workers.append(self.Process(target=self.fill_example_queue))
            self._example_q_workers[-1].daemon = True
            self._example_q_workers[-1].start()
          
        self._batch_q_workers = []
        for _ in range(self.num_batch_q_workers):
            self._batch_q_workers.append(self.Process(target=self.fill_batch_queue))
            self._batch_q_workers[-1].daemon = True
            self._batch_q_workers[-1].start()

        if not self.single_pass:
            self._watch_worker = self.Process(target=self.watch_workers)
            self._watch_worker.daemon = True
            self._watch_worker.start()
            
    
    # internal executive functions
    # example
    def fill_example_queue(self):
        """
        """
        example_iter = self.example_gen(single_pass=self.single_pass)
        #
        while True:
            try:
                base_example = next(example_iter)
                # print(base_example)
                #
            except BaseException: # if there is no more example:                
                if self.single_pass:
                    print("fill_example_queue(), single_pass on, data finished, break loop")
                    self._finished_reading = True                    
                    break
                else:
                    raise Exception("single_pass mode is off but the example generator is out of data; ERROR!")
            #
            self._example_queue.put(base_example)
            self.count_put_examples += 1
            
    # batch    
    def fill_batch_queue(self):
        """
        """
        while True:           
            if self.with_bucket:                    
                inputs = []
                for _ in range(self.batch_size * self.bucketing_cache_size):
                    inputs.append(self._example_queue.get(timeout = self.EXAMPLE_TIME_OUT))
                inputs = sorted(inputs, key = lambda item: item.seq_len) 
                
                batches = []
                for i in range(0, len(inputs), self.batch_size):
                    batches.append(inputs[i:i + self.batch_size])
                if not self.single_pass:
                    random.shuffle(batches)
                for b in batches:
                    self._batch_queue.put(self.batch_standardizer(b))
            else:
                b = []
                flag_succeed = 1
                for eid in range(self.batch_size):
                    try:
                        example = self._example_queue.get(timeout = self.EXAMPLE_TIME_OUT)
                        b.append(example)
                        self.count_get_examples += 1
                    except BaseException:
                        flag_succeed = 0
                        break
                #
                if flag_succeed == 0:
                    for example in b:
                        self._example_queue.put(example)
                    print("fill_batch_queue(), examples reput, num_examples: %d." % len(b))
                    break
                #
                self._batch_queue.put(self.batch_standardizer(b)) 
                # print(b)
                #
    
    # watch
    def watch_workers(self):
        """
        """        
        while True:
            time.sleep(60)
            if self._finished_reading: break
            #                    
            for idx, t in enumerate(self._example_q_workers):
                if not t.is_alive(): # if the worker is dead
                    print('found example queue worker dead. Restarting.')
                    new_t = self.Process(target=self.fill_example_queue)
                    self._example_q_workers[idx] = new_t
                    new_t.daemon = True
                    new_t.start()
            for idx, t in enumerate(self._batch_q_workers):
                if not t.is_alive(): # if the worker is dead
                    print('found batch queue worker dead. Restarting.')
                    new_t = self.Process(target=self.fill_batch_queue)
                    self._batch_q_workers[idx] = new_t
                    new_t.daemon = True
                    new_t.start()

# 
if __name__ == '__main__':
    
    pass




# -*- coding: utf-8 -*-

import os

from Zeras.vocab import Vocab
from model_settings import ModelSettings

import argparse

def parse_args():
    """
    Parses command line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--mode', choices=['train', 'eval', 'predict', 'convert'],
                        default = 'train', help = 'run mode')
    #
    parser.add_argument('--note', type=str, default = 'note_something',
                        help = 'make some useful notes')
    parser.add_argument('--debug', type=int, default = 0,
                        help = 'debug or not (using debug data or not)')
    parser.add_argument('--gpu', type=str, default = '0',
                        help = 'specify gpu device')
    #
    parser.add_argument('--ckpt_loading', choices=['best', 'latest'],
                        default = 'best', help='lastest ckpt or best')
    #
    parser.add_argument('--task', type=str, help = 'specify task',
                        default = 'copy')
    parser.add_argument('--settings', type=str, help='settings file',
                        default = None)
    parser.add_argument('--model_tag', type=str, help='model_tag',
                        default = 'transformer')
    #    
    return parser.parse_args()

#  
if __name__ == '__main__':
    
    args = parse_args()
    run_mode = args.mode
    #
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
    #
    # task & model
    task = args.task
    settings_file = args.settings
    if task == "copy":
        import task_copy_model_utils as model_utils
        if settings_file is None:
            settings_file = "./task_copy_settings.json"
        #
    #
    model_tag = args.model_tag
    if model_tag.startswith('transformer'):
        from model_graph_transformer import ModelGraph    
    #
    # settings
    settings = ModelSettings()
    settings.load_from_json_file(settings_file)
    settings.gpu_available = args.gpu
    settings.model_tag = model_tag
    #    
    if run_mode == 'predict':
        settings.is_train = False
    else:
        settings.is_train = True
    #
    settings.check_settings()
    settings.create_or_reset_log_file()
    settings.logger.info('running with args : {}'.format(args))
    settings.logger.info(settings.trans_info_to_dict())
    settings.save_to_json_file("./temp_settings.json")
    #
    # vocab
    vocab = Vocab()    
    vocab.add_tokens_from_file(settings.tokens_file)
    vocab.load_pretrained_embeddings(settings.emb_file)
    vocab.emb_dim = settings.emb_dim
    #
    # model & vocab
    settings.model_graph = ModelGraph
    settings.vocab = vocab
    #
    # run
    if run_mode == 'train':
        model_utils.do_train_and_valid(settings, args)
    elif run_mode == 'eval':
        model_utils.do_eval(settings, args)
    elif run_mode == 'predict':
        model_utils.do_predict(settings, args)
    elif run_mode == 'convert':
        model_utils.do_convert(settings, args)
    else:
        print('NOT supported mode. supported modes: train, eval, convert and predict.')
    #
    
    


# -*- coding: utf-8 -*-
"""
Created on Mon Aug 27 22:31:20 2018

@author: limingfan
"""

import os
import numpy as np

import tensorflow as tf
from tensorflow.python.framework import graph_util


"""
This class is meant to be task-agnostic.

"""

#
def get_warmup_and_exp_decayed_lr(settings, global_step):
    """ lr_base, warmup_steps, decay_steps, decay_rate, staircase
    """
    learning_rate = tf.constant(value = settings.learning_rate_base,
                                shape = [], dtype = tf.float32)    
    learning_rate = tf.train.exponential_decay(learning_rate, global_step,
                                               settings.decay_steps,
                                               settings.decay_rate,
                                               settings.staircase)    
    if settings.warmup_steps:
        global_steps_int = tf.cast(global_step, tf.int32)
        warmup_steps_int = tf.constant(settings.warmup_steps, dtype=tf.int32)
        
        global_steps_float = tf.cast(global_steps_int, tf.float32)
        warmup_steps_float = tf.cast(warmup_steps_int, tf.float32)
        
        warmup_percent_done = global_steps_float / warmup_steps_float
        warmup_learning_rate = settings.learning_rate_base * warmup_percent_done
        
        learning_rate = tf.cond(global_steps_int < warmup_steps_int,
                                lambda: warmup_learning_rate,
                                lambda: learning_rate)
    #
    return learning_rate
    #
    
#
class ModelWrapper():
    
    def __init__(self, settings, model_graph,
                 learning_rate_schedule = None, customized_optimizer = None):
        #
        self.set_model_settings(settings)
        self.model_graph = model_graph
        #
        if learning_rate_schedule is None:
            self.learning_rate_schedule = get_warmup_and_exp_decayed_lr
        else:
            self.learning_rate_schedule = learning_rate_schedule
        #
        self.customized_optimizer = customized_optimizer
        # self._opt = self.customized_optimizer(self.settings, self.learning_rate_tensor)
        #
        
    def set_model_settings(self, settings):
        #
        self.settings = settings
        #
        for key in settings.__dict__.keys():                 
            self.__dict__[key] = settings.__dict__[key]
        #
        self.num_gpu = len(self.gpu_available.split(","))
        #
        # session info
        self.sess_config = tf.ConfigProto(log_device_placement = settings.log_device,
                                          allow_soft_placement = settings.soft_placement)
        self.sess_config.gpu_options.allow_growth = settings.gpu_mem_growth
        #
    
    # predict
    def prepare_for_prediction(self, pb_file_path = None):
        #
        if pb_file_path is None: pb_file_path = self.pb_file 
        if not os.path.exists(pb_file_path):
            assert False, 'ERROR: %s NOT exists, when prepare_for_prediction()' % pb_file_path
        #
        self._graph = tf.Graph()
        with self._graph.as_default():
            with open(pb_file_path, "rb") as f:
                graph_def = tf.GraphDef()
                graph_def.ParseFromString(f.read())
                tf.import_graph_def(graph_def, name="")
                #
            #
            self._inputs_predict = []
            for item in self.inputs_predict_name:
                tensor = self._graph.get_tensor_by_name(item)
                self._inputs_predict.append(tensor)
            #
            self._outputs_predict = []
            for item in self.outputs_predict_name:
                tensor = self._graph.get_tensor_by_name(item)
                self._outputs_predict.append(tensor)
            #
            self._inputs_predict_num = len(self._inputs_predict)
            print('Graph loaded for prediction')
            #
        #
        self._sess = tf.Session(graph = self._graph, config = self.sess_config)
        #
    
    def predict_from_batch(self, x_batch):
        #
        feed_dict = self.feed_data_predict(x_batch)
        outputs = self._sess.run(self._outputs_predict, feed_dict = feed_dict)
        
        return outputs

    def feed_data_predict(self, x_batch):        
        feed_dict = {}
        for idx in range(self._inputs_predict_num):
            feed_dict[self._inputs_predict[idx] ] = x_batch[idx]
        return feed_dict
        
    def feed_data_train(self, data_batch):        
        feed_dict = {}
        for idx in range(self._inputs_train_num):
            feed_dict[self._inputs_train[idx] ] = data_batch[idx]
        return feed_dict
    
    # one_batch functions
    def run_train_one_batch(self, one_batch):
        #
        feed_dict = self.feed_data_train(one_batch)
        loss, lr, _ = self._sess.run([self._loss_tensor, self.learning_rate_tensor,
                                      self._train_op], feed_dict = feed_dict)
        return loss, lr
        
    def run_eval_one_batch(self, one_batch):
        #
        feed_dict = self.feed_data_train(one_batch)        
        metric = None
        if self.use_metric:         
            *results, loss, metric = self._sess.run(self._outputs_eval,
                                                    feed_dict = feed_dict)
        else:
            *results, loss = self._sess.run(self._outputs_eval,
                                            feed_dict = feed_dict)         
        return results, loss, metric
        
    def run_debug_one_batch(self, one_batch):
        #
        assert self.num_gpu == 1, "debug mode can only be run with single gpu"
        #
        feed_dict = self.feed_data_train(one_batch)
        results = self._sess.run(self._debug_tensors, feed_dict = feed_dict)        
        return results

    #
    def prepare_for_train_and_valid(self, dir_ckpt = None):
        #
        if self.num_gpu == 1:
            self.prepare_for_train_and_valid_single_gpu(dir_ckpt)
        else:
            self.prepare_for_train_and_valid_multi_gpu(dir_ckpt)
        #
    
    #
    def prepare_for_train_and_valid_single_gpu(self, dir_ckpt = None):

        # graph
        self._graph = tf.Graph()
        with self._graph.as_default():
            #
            self.global_step = tf.get_variable("global_step", shape=[], dtype=tf.int32,
                                               initializer = tf.constant_initializer(0),
                                               trainable = False)
            #
            self.learning_rate_tensor = self.learning_rate_schedule(self.settings, self.global_step)
            self.learning_rate_tensor = tf.identity(self.learning_rate_tensor, name= "lr")
            #
            # optimizer
            # optimizer = tf.train.MomentumOptimizer(learning_rate, MOMENTUM, use_nesterov=True)
            # optimizer = tf.train.AdadeltaOptimizer(learning_rate=self.lr, epsilon=1e-6)              
            # optimizer = tf.train.AdamOptimizer(learning_rate = self.learning_rate, beta1 = MOMENTUM)
            #
            if self.optimizer_type == 'sgd':
                self._opt = tf.train.GradientDescentOptimizer(self.learning_rate_tensor)
            elif self.optimizer_type == 'momentum':
                self._opt = tf.train.MomentumOptimizer(self.learning_rate_tensor, self.momentum, use_nesterov=True)
            elif self.optimizer_type == 'adam':
                self._opt = tf.train.AdamOptimizer(learning_rate = self.learning_rate_tensor, beta1 = self.momentum)
            elif self.optimizer_type == 'customized':
                self._opt = self.customized_optimizer(self.settings, self.learning_rate_tensor)
            else:
                assert False, "NOT supported optimizer_type"
            #
            # model
            input_tensors, label_tensors = self.model_graph.build_placeholder(self.settings)
            #
            vs_str = self.vs_str_multi_gpu
            vs_prefix = vs_str + "/"
            with tf.variable_scope(vs_str):
                output_tensors = self.model_graph.build_inference(self.settings, input_tensors)
                loss, metric = self.model_graph.build_loss_and_metric(self.settings, output_tensors, label_tensors)
            #
            # all trainable vars
            self.trainable_vars = tf.trainable_variables()
            # print(self.trainable_vars)
            #
            # keep_prob
            self._keep_prob = self._graph.get_tensor_by_name(vs_prefix + "keep_prob:0")
            #
            # metric and loss
            if self.use_metric:
                self._metric_tensor = metric # self._graph.get_tensor_by_name(self.metric_name)
            #
            self._loss_tensor = loss # self._graph.get_tensor_by_name(self.loss_name)
            #
            # regularization
            def is_excluded(v):
                for item in self.reg_exclusions:
                    if item in v.name: return True
                return False
            #
            if self.reg_lambda > 0.0:
                loss_reg = tf.add_n( [tf.nn.l2_loss(v) for v in self.trainable_vars
                                     if not is_excluded(v)] )
                loss_reg = tf.multiply(loss_reg, self.reg_lambda)
                self._loss_tensor = tf.add(self._loss_tensor, loss_reg)
            #
            # grad_clip
            grad_and_vars = self._opt.compute_gradients(self._loss_tensor)            
            if self.grad_clip > 0.0:
                gradients, variables = zip(*grad_and_vars)
                grads, _ = tf.clip_by_global_norm(gradients, self.grad_clip)
                grad_and_vars = zip(grads, variables)
            #
            # train_op
            self._train_op = self._opt.apply_gradients(grad_and_vars,
                                                       global_step = self.global_step)
            #                 
            # save info
            self._saver = tf.train.Saver()
            self._saver_best = tf.train.Saver()
            
            # sess
            self._sess = tf.Session(graph=self._graph, config = self.sess_config)

            # initialize the model
            self._sess.run(tf.global_variables_initializer())
            self.assign_dropout_keep_prob(self.keep_prob)
            # self.assign_learning_rate(self.learning_rate_base)
            
            # params count
            self.num_vars = len(self.trainable_vars)
            str_info = 'graph built, there are %d variables in the model' % self.num_vars
            self.logger.info(str_info)
            print(str_info)
            #
            tf_shapes = [tf.shape(v) for v in self.trainable_vars]
            shapes_v = self._sess.run(tf_shapes)
            params_v = [np.prod(item) for item in shapes_v]
            self.param_num = sum(params_v)
            #
            str_info = 'there are %d parameters in the model' % self.param_num
            self.logger.info(str_info)
            print(str_info)
            #
            print()
            for idx in range(self.num_vars):
                print(self.trainable_vars[idx])
                print(params_v[idx])
            print()
            #
            
            # outputs train
            self._outputs_train = output_tensors
            #
            # outputs eval
            self._outputs_eval = []
            for item in self._outputs_train:
                self._outputs_eval.append(item)
            self._outputs_eval.append(self._loss_tensor)
            #
            if self.use_metric:
                self._outputs_eval.append(self._metric_tensor)
            #
            # train inputs
            self._inputs_train = []
            for item in self.inputs_train_name:
                tensor = self._graph.get_tensor_by_name(item)
                self._inputs_train.append(tensor)
            #          
            self._inputs_train_num = len(self._inputs_train)
            #
            # debug tensors
            self._debug_tensors = []
            for name in self.debug_tensors_name:
                tensor = self._graph.get_tensor_by_name(name)
                self._debug_tensors.append(tensor)
            #

        #
        # load
        if dir_ckpt is None: dir_ckpt = self.model_dir + '_best'
        self.load_ckpt(dir_ckpt)
        #
        
    #
    @staticmethod
    def sum_up_gradients(list_grad_bundles):
        """ list_grad_bundles: [ [(g1,v1), (g2, v2), ...],
                                 [(g1,v1), (g2, v2), ...], ...,
                                 [(g1,v1), (g2, v2), ...] ]
            zip(*list_grad_bundles): [ ... ]
        """
        summed_grads = []
        for grads_per_var in zip(*list_grad_bundles):
            grads = []
            for g, _ in grads_per_var:
                expanded_g = tf.expand_dims(g, 0)
                grads.append(expanded_g)
            #
            grads_concat = tf.concat(grads, 0)
            grads_sum = tf.reduce_sum(grads_concat, 0)
            grad_and_var = (grads_sum, grads_per_var[0][1])
            summed_grads.append(grad_and_var)
        #
        return summed_grads
    
    #
    def prepare_for_train_and_valid_multi_gpu(self, dir_ckpt = None):
        
        gpu_batch_split = self.gpu_batch_split

        # graph
        self._graph = tf.Graph()
        with self._graph.as_default():
            #
            self.global_step = tf.get_variable("global_step", shape=[], dtype=tf.int32,
                                               initializer = tf.constant_initializer(0),
                                               trainable = False)
            #
            self.learning_rate_tensor = self.learning_rate_schedule(self.settings, self.global_step)
            self.learning_rate_tensor = tf.identity(self.learning_rate_tensor, name= "lr")
            #
            # optimizer
            # optimizer = tf.train.MomentumOptimizer(learning_rate, MOMENTUM, use_nesterov=True)
            # optimizer = tf.train.AdadeltaOptimizer(learning_rate=self.lr, epsilon=1e-6)              
            # optimizer = tf.train.AdamOptimizer(learning_rate = self.learning_rate, beta1 = MOMENTUM)
            #
            if self.optimizer_type == 'sgd':
                self._opt = tf.train.GradientDescentOptimizer(self.learning_rate_tensor)
            elif self.optimizer_type == 'momentum':
                self._opt = tf.train.MomentumOptimizer(self.learning_rate_tensor, self.momentum, use_nesterov=True)
            elif self.optimizer_type == 'adam':
                self._opt = tf.train.AdamOptimizer(learning_rate = self.learning_rate_tensor, beta1 = self.momentum)
            elif self.optimizer_type == 'customized':
                self._opt = self.customized_optimizer(self.settings, self.learning_rate_tensor)
            else:
                assert False, "NOT supported optimizer_type"
            #
            # model, placeholder
            input_tensors, label_tensors = self.model_graph.build_placeholder(self.settings)
            #
            # split among gpu
            inputs = []
            labels = []
            #
            for idx in range(len(input_tensors)):
                in_split = tf.split(input_tensors[idx], gpu_batch_split, axis = 0)
                inputs.append(in_split)
            #
            for idx in range(len(label_tensors)):
                lb_split = tf.split(label_tensors[idx], gpu_batch_split, axis = 0)
                labels.append(lb_split)
            #
            inputs_split = list(zip(*inputs))
            labels_split = list(zip(*labels))
            #
            # model, inference, loss
            outputs_list = []
            loss_list = []
            metric_list = []
            grads_bundles = []
            #
            vs_str = self.vs_str_multi_gpu
            vs_prefix = vs_str + "/"
            with tf.variable_scope(vs_str):
                for gid in range(self.num_gpu):
                    with tf.device("/gpu:%d" % gid), tf.name_scope("bundle_%d" % gid):
                        #
                        output_tensors = self.model_graph.build_inference(self.settings,
                                                                          inputs_split[gid])
                        loss, metric = self.model_graph.build_loss_and_metric(self.settings,
                                                                              output_tensors,
                                                                              labels_split[gid])
                        #
                        tf.get_variable_scope().reuse_variables()
                        #
                        grads = self._opt.compute_gradients(loss)
                        grads_bundles.append(grads)
                        #
                        outputs_list.append(output_tensors)
                        loss_list.append(loss)
                        metric_list.append(metric)
                        #
            #
            # all trainable vars
            self.trainable_vars = tf.trainable_variables()
            # print(self.trainable_vars)
            #
            # keep_prob
            self._keep_prob = self._graph.get_tensor_by_name(vs_prefix + "keep_prob:0")
            #
            # regularization
            def is_excluded(v):
                for item in self.reg_exclusions:
                    if item in v.name: return True
                return False
            #
            if self.reg_lambda > 0.0:
                loss_reg = tf.add_n( [tf.nn.l2_loss(v) for v in self.trainable_vars
                                     if not is_excluded(v)] )
                loss_reg = tf.multiply(loss_reg, self.reg_lambda)
                grads_reg = self._opt.compute_gradients(loss_reg)
                #
                grads_reg_clear = []
                for g, v in grads_reg:
                    if g is None: g = tf.zeros_like(v)
                    grads_reg_clear.append( (g,v) )
                    #
                #
                grads_bundles.append(grads_reg_clear)
                #
            #           
            # grad sum
            grads_summed = ModelWrapper.sum_up_gradients(grads_bundles)
            #            
            # grad_clip
            if self.grad_clip > 0.0:
                gradients, variables = zip(*grads_summed)
                grads, _ = tf.clip_by_global_norm(gradients, self.grad_clip)
                grads_summed = zip(grads, variables)
            #
            # train_op
            self._train_op = self._opt.apply_gradients(grads_summed,
                                                       global_step = self.global_step)
            #               
            # save info
            self._saver = tf.train.Saver()
            self._saver_best = tf.train.Saver()
            
            # sess
            self._sess = tf.Session(graph=self._graph, config = self.sess_config)

            # initialize the model
            self._sess.run(tf.global_variables_initializer())
            self.assign_dropout_keep_prob(self.keep_prob)
            # self.assign_learning_rate(self.learning_rate_base)
            
            # params count
            self.num_vars = len(self.trainable_vars)
            str_info = 'graph built, there are %d variables in the model' % self.num_vars
            self.logger.info(str_info)
            print(str_info)
            #
            tf_shapes = [tf.shape(v) for v in self.trainable_vars]
            shapes_v = self._sess.run(tf_shapes)
            params_v = [np.prod(item) for item in shapes_v]
            self.param_num = sum(params_v)
            #
            str_info = 'there are %d parameters in the model' % self.param_num
            self.logger.info(str_info)
            print(str_info)
            #
            print()
            for idx in range(self.num_vars):
                print(self.trainable_vars[idx])
                print(params_v[idx])
            print()
            #
            
            #
            # loss
            loss_w = [loss_list[gid] * self.gpu_batch_split[gid]
                                         for gid in range(self.num_gpu)]
            self._loss_tensor = tf.add_n(loss_w) / self.batch_size
            #
            # metric
            if self.use_metric:
                metric_w = [metric_list[gid] * self.gpu_batch_split[gid]
                                                for gid in range(self.num_gpu)]
                self._metric_tensor = tf.add_n(metric_w) / self.batch_size
            #
            # output
            outputs_list_z = zip(*outputs_list)
            outputs_list_c = []
            for item in outputs_list_z:
                item_c = tf.concat(item, 0)
                outputs_list_c.append(item_c)
            #
            self._outputs_train = outputs_list_c
            #
            # outputs eval
            if self.use_metric:
                self._outputs_eval = self._outputs_train + [self._loss_tensor,
                                                            self._metric_tensor]
            else:
                self._outputs_eval = self._outputs_train + [self._loss_tensor]
            #
            # train inputs
            self._inputs_train = []
            for item in self.inputs_train_name:
                tensor = self._graph.get_tensor_by_name(item)
                self._inputs_train.append(tensor)
            #          
            self._inputs_train_num = len(self._inputs_train)
            #
            
        #
        # load
        if dir_ckpt is None: dir_ckpt = self.model_dir + '_best'
        self.load_ckpt(dir_ckpt)
        #
        
    #
    def assign_dropout_keep_prob(self, keep_prob):
        #
        with self._graph.as_default():
            self._sess.run(tf.assign(self._keep_prob, tf.constant(keep_prob, dtype=tf.float32)))
        #
    
    #
    def load_ckpt_and_save_pb_file(self, dir_ckpt):
        #
        is_train = self.settings.is_train
        self.settings.is_train = False       #
        #
        model = ModelWrapper(self.settings, self.settings.model_graph)
        model.prepare_for_train_and_valid_single_gpu(dir_ckpt)         # loaded here 
        model.assign_dropout_keep_prob(1.0)
        #
        pb_file = os.path.join(dir_ckpt, "model_saved.pb")
        #
        constant_graph = graph_util.convert_variables_to_constants(
                model._sess, model._sess.graph_def,
                output_node_names = model.pb_outputs_name)
        with tf.gfile.FastGFile(pb_file, mode='wb') as f:
            f.write(constant_graph.SerializeToString())
        #
        str_info = 'pb_file saved: %s' % pb_file
        self.logger.info(str_info)
        #
        self.settings.is_train = is_train           #
        #
            
    def save_ckpt_best(self, model_dir, model_name, step):
        #
        self._saver_best.save(self._sess, os.path.join(model_dir, model_name),
                              global_step = step)
        
    def save_ckpt(self, model_dir, model_name, step):
        #
        self._saver.save(self._sess, os.path.join(model_dir, model_name),
                         global_step = step)
    
    def load_ckpt(self, dir_ckpt):
        #
        ckpt = tf.train.get_checkpoint_state(dir_ckpt)        
        if ckpt and ckpt.model_checkpoint_path:
            self._saver.restore(self._sess, ckpt.model_checkpoint_path)
            #
            str_info = 'ckpt loaded from %s' % dir_ckpt
            self.logger.info(str_info)
            print(str_info)
        else:
            str_info = 'Failed: ckpt loading from %s' % dir_ckpt
            self.logger.info(str_info)
            print(str_info)            
    
    # graph and sess
    def get_model_graph_and_sess(self):
        #
        return self._graph, self._sess
        #
            
if __name__ == '__main__':
    
    pass

        
    
    


# -*- coding: utf-8 -*-

from __future__ import with_statement

import sys
if sys.version_info < (3, 5):
    sys.exit('Python 3.5 or greater is required.')

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import Zeras


with open('README.md') as fp:
    readme = fp.read()

with open('LICENSE') as fp:
    license = fp.read()

setup(name = 'Zeras',
      version = Zeras.__version__,
      description = 'A simple tensorflow wrapper.',
      long_description = readme,
      author = 'Ming-Fan Li',
      author_email = 'li_m_f@163.com',
      maintainer='Ming-Fan Li',
      maintainer_email='li_m_f@163.com',
      url='https://github.com/Li-Ming-Fan/Zeras',
      packages=['Zeras'],
      license=license,
      platforms=['any'],
      classifiers=[]
      )



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 28 21:12:23 2018

@author: limingfan
"""

import os
import time
import json

import logging


class ModelSettingsBaseboard(object):
    """
    """
    def __init__(self):

        # model
        self.model_tag = None     # str
        self.is_train = None      # bool
        
        # model graph macro parameters
        #
        
        # other macro parameters
        #
        
        #
        # train
        self.gpu_available = "0"      # could be specified in args
        self.gpu_batch_split = None   # list, for example, [12, 20]; if None, batch split evenly
        #
        self.gpu_mem_growth = True
        self.log_device = False
        self.soft_placement = True

        self.num_epochs = 100     
        self.batch_size = 32
        self.batch_size_eval = 6
        self.max_batches_eval = 20
        
        self.reg_lambda = 0.001  # 0.0, 0.01
        self.reg_exclusions = ["embedding", "bias", "layer_norm", "LayerNorm"]
        self.grad_clip = 8.0  # 0.0, 5.0, 8.0, 2.0
        self.keep_prob = 0.8  # 1.0, 0.7, 0.5
        self.label_smoothing = 0.01
        
        self.optimizer_type = 'adam'  # adam, momentum, sgd, customized
        self.momentum = 0.9
        self.learning_rate_base = 0.001   #        
        self.learning_rate_minimum = 0.000001
        self.warmup_steps = 1000
        self.decay_steps = 1000
        self.decay_rate = 0.99
        self.staircase = True
        
        self.check_period_batch = 100
        self.valid_period_batch = 100
        #

        # inputs/outputs
        self.vs_str_multi_gpu = "vs_multi_gpu"
        #
        self.inputs_predict_name = ['src_seq:0', 'src_seq_mask:0']
        self.outputs_predict_name = ['vs_multi_gpu/logits:0']
        self.pb_outputs_name = ['vs_multi_gpu/logits']
                
        self.inputs_train_name = ['src_seq:0', 'src_seq_mask:0',
                                  'dcd_seq:0', 'dcd_seq_mask:0',
                                  'labels_seq:0', 'labels_mask:0']
        self.outputs_train_name = ['vs_multi_gpu/logits:0']
        self.use_metric = True
        
        self.debug_tensors_name = ['vs_multi_gpu/loss/loss:0',
                                   'vs_multi_gpu/logits:0',
                                   'vs_multi_gpu/preds:0'
                                   #'encoder/encoder_rnn_1/bw/bw/sequence_length:0',
                                   #'inputs_len:0'
                                   ]
        
        #
        # save and log, if not set, default values will be used.
        self.base_dir = './task_results'
        self.model_dir = None
        self.model_name = None
        self.pb_file = None
        self.log_dir = None
        self.log_path = None
        #
    
    def check_settings(self):
        """ assert and make directories
        """        
        # assert         
        assert self.is_train is not None, 'is_train not assigned'
        
        # gpu
        num_gpu = len(self.gpu_available.split(","))
        if num_gpu > 1:
            #
            if self.gpu_batch_split is None:
                self.gpu_batch_split = [self.batch_size//num_gpu] * num_gpu
            #
            str_info = "make sure that num_gpu == len(self.gpu_batch_split)"
            assert num_gpu == len(self.gpu_batch_split), str_info
            str_info = "make sure that self.batch_size == sum(self.gpu_batch_split)"
            assert self.batch_size == sum(self.gpu_batch_split), str_info
            
        # directories
        if self.model_dir is None:
            self.model_dir = os.path.join(self.base_dir, 'model_' + self.model_tag)
        if self.log_dir is None: self.log_dir = os.path.join(self.base_dir, 'log')
        #
        if not os.path.exists(self.base_dir): os.mkdir(self.base_dir)
        if not os.path.exists(self.model_dir): os.mkdir(self.model_dir)
        if not os.path.exists(self.model_dir + '_best'): os.mkdir(self.model_dir + '_best')
        if not os.path.exists(self.log_dir): os.mkdir(self.log_dir)
        #
        # files
        if self.model_name is None: self.model_name = 'model_' + self.model_tag
        if self.pb_file is None: self.pb_file = os.path.join(self.model_dir + '_best',
                                                             'model_saved.pb')
        #
        # logger
        str_datetime = time.strftime("%Y-%m-%d-%H-%M")       
        if self.log_path is None: self.log_path = os.path.join(
                self.log_dir, self.model_name + "_" + str_datetime +".txt")
        #
        self.logger = logging.getLogger(self.log_path)  # use log_path as log_name
        self.logger.setLevel(logging.INFO)
        handler = logging.FileHandler(self.log_path)
        handler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        # formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        # self.logger.info('test')
        #
        self.display()
        #
        
    def create_or_reset_log_file(self):        
        with open(self.log_path, 'w', encoding='utf-8'):
            pass
        
    def close_logger(self):
        for item in self.logger.handlers: item.close()
    
    #
    def display(self):
        """
        """        
        print()
        for name,value in vars(self).items():
            if not isinstance(value, (int, float, str, bool, list, dict, tuple)):
                continue
            print(str(name) + ': ' + str(value))
        print()
    
    #
    def trans_info_to_dict(self):
        """
        """                
        info_dict = {}
        for name,value in vars(self).items():
            if not isinstance(value, (int, float, str, bool, list, dict, tuple)):
                continue
            info_dict[str(name)] = value        
        return info_dict
    
    def assign_info_from_dict(self, info_dict):
        """
        """
        for key in info_dict:
            value = info_dict[key]
            setattr(self, key, value)
        #
        
    def save_to_json_file(self, file_path):
        """
        """
        info_dict = self.trans_info_to_dict()
        #
        info_dict["model_dir"] = None
        info_dict["model_name"] = None
        info_dict["pb_file"] = None
        info_dict["log_dir"] = None
        info_dict["log_path"] = None
        #
        with open(file_path, "w", encoding="utf-8") as fp:
            json.dump(info_dict, fp, ensure_ascii=False, indent=4)
        #       
    
    def load_from_json_file(self, file_path):
        """
        """
        if file_path is None:
            print("model settings file: %s NOT found, using default settings"
                  % file_path)
            return
        #
        with open(file_path, "r", encoding="utf-8") as fp:
            info_dict = json.load(fp)
        #
        self.assign_info_from_dict(info_dict)
        #
        
        
    
        
if __name__ == '__main__':
    
    sett = ModelSettingsBaseboard()
    
    sett.model_tag = 'cnn'
    sett.is_train = False
    
    #print(dir(sett))
    #l = [i for i in dir(sett) if inspect.isbuiltin(getattr(sett, i))]
    #l = [i for i in dir(sett) if inspect.isfunction(getattr(sett, i))]
    #l = [i for i in dir(sett) if not callable(getattr(sett, i))]
    
    sett.check_settings()
    
    print(sett.__dict__.keys())
    print()
    
    info_dict = sett.trans_info_to_dict()
    print(info_dict)
    print()
    
    #
    info_dict["model_tag"] = "transformer"
    sett.assign_info_from_dict(info_dict)
    
    #
    info_dict = sett.trans_info_to_dict()
    print(info_dict)
    print()
    #
    
    file_path = "./settings_template.json"
    sett.save_to_json_file(file_path)    
    sett.load_from_json_file(file_path)
    
    #
    info_dict = sett.trans_info_to_dict()
    print(info_dict)
    print()
    #

    #    
    sett.close_logger()
    #


# -*- coding: utf-8 -*-
"""
Created on Thu Jan 31 05:08:20 2019

@author: limingfan
"""

import os
import numpy as np

from Zeras.data_batcher import DataBatcher
from Zeras.model_wrapper import ModelWrapper

import task_copy_data_set as data_set

#
def eval_process(model, eval_batcher, max_batches_eval, mode_eval):
    """
    """
    loss_aver, metric_aver = 0.0, 0.0
    count = 0
    while True:
        batch = eval_batcher.get_next_batch()  
        #
        if batch is None: break
        if count == max_batches_eval: continue  #
        #
        count += 1
        # print(count)
        #
        results, loss, metric = model.run_eval_one_batch(batch)
        loss_aver += loss
        metric_aver += metric
        # print(loss)
        # print(metric)
        #
        if mode_eval:
            print(count)
            print("batch data:")
            print(batch[4])
            #
            print("results:")
            print(np.argmax(results[0], -1) )
            print()        
        #
    #
    loss_aver /= count
    # metric_aver /= count
    metric_aver = 100 - loss_aver
    #
    model.logger.info('eval finished, with total num_batches: %d' % count)
    # model.logger.info('loss_aver, metric_aver: %g, %g' % (loss_aver, metric_aver))
    #
    eval_score = {}
    return eval_score, loss_aver, metric_aver
    #
    
def do_eval(settings, args):
    #
    if args.ckpt_loading == "latest":
        dir_ckpt = settings.model_dir
    else:
        dir_ckpt = settings.model_dir + "_best"
    #
    # model
    model = ModelWrapper(settings, settings.model_graph)
    model.prepare_for_train_and_valid(dir_ckpt)
    model.assign_dropout_keep_prob(1.0)
    #
    # data
    data_batcher = DataBatcher(data_set.get_examples_generator(),
                               data_set.batch_std_transor,
                               settings.batch_size_eval, single_pass = True,
                               worker_type="thread")
    #
    # eval
    eval_score, loss_aver, metric_aver = eval_process(model, data_batcher,
                                                      settings.max_batches_eval,
                                                      mode_eval = True)
    #
    print('loss_aver, metric_aver: %g, %g' % (loss_aver, metric_aver))
    model.logger.info('loss_aver, metric_aver: %g, %g' % (loss_aver, metric_aver))
    model.logger.info('{}'.format(eval_score))
    #
    
def do_train_and_valid(settings, args):
    #
    if args.ckpt_loading == "latest":
        dir_ckpt = settings.model_dir
    else:
        dir_ckpt = settings.model_dir + "_best"
    #
    # model
    model = ModelWrapper(settings, settings.model_graph,
                         learning_rate_schedule = None,
                         customized_optimizer = None)
    model.prepare_for_train_and_valid(dir_ckpt)
    #    
    # data
    data_batcher = DataBatcher(data_set.get_examples_generator(),
                               data_set.batch_std_transor,
                               settings.batch_size, single_pass = False,
                               worker_type="thread")
    eval_period = settings.valid_period_batch
    #
    # train
    loss = 10000.0
    best_metric_val = 0
    # last_improved = 0
    lr = 0.0
    #
    count = 0
    model.logger.info("")
    while True:
        #
        # eval
        if count % eval_period == 0:            
            model.logger.info("training curr batch, loss, lr: %d, %g, %g" % (count, loss, lr) )
            #
            model.save_ckpt(model.model_dir, model.model_name, count)
            model.assign_dropout_keep_prob(1.0)
            #
            model.logger.info('evaluating after num_batches: %d' % count)
            eval_batcher = DataBatcher(data_set.get_examples_generator(),
                                       data_set.batch_std_transor,
                                       settings.batch_size, single_pass = True,
                                       worker_type="thread")
            #
            eval_score, loss_aver, metric_val = eval_process(model, eval_batcher,
                                                             settings.max_batches_eval,
                                                             mode_eval = False)
            model.logger.info("eval loss_aver, metric, metric_best: %g, %g, %g" % (
                    loss_aver, metric_val, best_metric_val) )
            #
            # save best
            if metric_val >= best_metric_val:  # >=
                best_metric_val = metric_val
                # last_improved = count
                # ckpt
                model.logger.info('a new best model, saving ...')
                model.save_ckpt_best(model.model_dir + '_best', model.model_name, count)
                #
            """
            # decay
            if count - last_improved >= model.patience_decay:
                lr *= model.ratio_decay
                model.assign_learning_rate(lr)
                last_improved = count
                model.logger.info('learning_rate decayed after num_batches: %d' % count)
                model.logger.info('current learning_rate %g' % lr)
                #
                if lr < model.learning_rate_minimum:
                    model.logger.info('current learning_rate < learning_rate_minimum, stop training')
                    break
                #
            """
            #
            # lr *= model.ratio_decay
            # model.assign_learning_rate(lr)
            # model.logger.info('learning_rate decayed after num_batches: %d' % count)
            # model.logger.info('current learning_rate: %g' % lr)
            #
            if lr < model.learning_rate_minimum and count > settings.warmup_steps:
                model.logger.info('current learning_rate < learning_rate_minimum, stop training')
                break
            #
            model.assign_dropout_keep_prob(settings.keep_prob)
            model.logger.info("")
            #
        #
        # train
        batch = data_batcher.get_next_batch()  
        # if batch is None: break
        count += 1
        # print(count)        
        #
        loss, lr = model.run_train_one_batch(batch)   # just for train
        # print(loss)
        # model.logger.info("training curr batch, loss, lr: %d, %g, %g" % (count, loss, lr)
        #
    #
    model.logger.info("training finshed with total num_batches: %d" % count)
    #
    
def do_predict(settings, args):
    #
    if args.ckpt_loading == "latest":
        dir_ckpt = settings.model_dir
    else:
        dir_ckpt = settings.model_dir + "_best"
    #
    pb_file = os.path.join(dir_ckpt, "model_saved.pb")
    #
    # model
    model = ModelWrapper(settings, settings.model_graph)
    model.prepare_for_prediction(pb_file)    
    #
    # data
    data_batcher = DataBatcher(data_set.get_examples_generator(),
                               data_set.batch_std_transor,
                               settings.batch_size_eval, single_pass = True,
                               worker_type="thread")
    #
    # predict
    count = 0
    while True:
        batch = data_batcher.get_next_batch()  
        #
        if batch is None: break
        if count == settings.max_batches_eval: continue  #
        #
        count += 1
        print(count)
        #
        print("batch data:")
        print(batch[4])
        print("batch data end")
        #
        results = model.predict_from_batch(batch)[0]
        #
        print("results:")
        print(np.argmax(results, -1) )
        print("results end")
        print()
        #
    #
    model.logger.info('prediction finished, with total num_batches: %d' % count)
    #

def do_convert(settings, args):
    #
    if args.ckpt_loading == "latest":
        dir_ckpt = settings.model_dir
    else:
        dir_ckpt = settings.model_dir + "_best"
    #
    # pb_file = os.path.join(dir_ckpt, "model_saved.pb")
    #
    # model
    model = ModelWrapper(settings, settings.model_graph)
    model.load_ckpt_and_save_pb_file(dir_ckpt)
    #